import os

import pandas as pd
import numpy as np
import plotly.graph_objects as go


class MajorContamination():
    """
    Major contamination.
    """

    def __init__(self, threshold):
        self.threshold = threshold

    def to_dataframe(self, samples):

        data = pd.DataFrame(
            columns=['sample_name', 'sample_group', 'sample_sex', 'sample_type',
                     'total_sites', 'total_heterozygous_sites',
                     'major_contamination'])

        for sample_name, sample in samples.items():

            row = {
                'sample_name': sample.sample_name,
                'sample_group': sample.sample_group,
                'sample_sex': sample.sample_sex,
                'sample_type': sample.sample_type,
                'total_sites': sample.metrics['major_contamination']['total_sites'],
                'total_heterozygous_sites': sample.metrics['major_contamination']['total_heterozygous_sites'],
                'major_contamination': sample.metrics['major_contamination']['val']
            }

            data = data.append(row, ignore_index=True)

        data = data.sort_values('major_contamination', ascending=False)
        return data

    def plot(self, samples, outdir):
        """
        Plot minor contamination data.
        """

        data = self.to_dataframe(samples)
        data['major_contamination'] = data['major_contamination'].map(
            lambda x: round(x, 5))

        fig = go.Figure()
        fig.add_trace(
            go.Bar(
                x=data['sample_name'],
                y=data['major_contamination'],
                customdata=data.to_numpy(),
                hovertemplate='<b>Sample group:</b> %{customdata[1]}' +
                              '<br><b>Sample name:</b> %{customdata[0]}' +
                              '<br><b>Sample sex:</b> %{customdata[2]}' +
                              '<br><b>Sample type:</b> %{customdata[3]}' +
                              '<br><b>Total sites:</b> %{customdata[5]}' +
                              '<br><b>Total heterozygous sites:</b> %{customdata[6]}' +
                              '<br><b>Major contamination:</b> %{y:E}' +
                              '<extra></extra>',
            ))
        fig.update_layout(
            yaxis_title="Major contamination",
            title_text="Major contamination across samples")
        fig.add_hline(y=self.threshold, line_color='red')

        fig.write_html(os.path.join(outdir, 'major_contamination.html'))

    def estimate(self, samples):
        """
        Estimate major contamination.
        """

        for sample_name, sample in samples.items():

            sites = sample.pileup
            sites_notna = sites[~pd.isna(sites['genotype_class'])]

            het_sites = sites_notna[sites_notna['genotype_class'] == 'Het']

            sample.metrics['major_contamination'] = {}
            sample.metrics['major_contamination']['total_sites'] = len(sites_notna)
            sample.metrics['major_contamination']['total_heterozygous_sites'] = len(het_sites)

            if sample.metrics['major_contamination']['total_sites'] == 0:
                sample.metrics['major_contamination']['val'] = np.nan
            else:
                sample.metrics['major_contamination']['val'] = \
                    len(het_sites) / len(sites_notna)

        return samples



#!/usr/bin/env python

"""Tests for `biometrics` package."""


import os
import argparse
from unittest import TestCase
from unittest import mock

from biometrics.biometrics import get_samples, run_minor_contamination, run_major_contamination
from biometrics.cli import get_args
from biometrics.extract import Extract
from biometrics.genotype import Genotyper
from biometrics.sex_mismatch import SexMismatch

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


class TestBiometrics(TestCase):
    """Tests for `biometrics` package."""

    @mock.patch(
        'argparse.ArgumentParser.parse_args',
        return_value=argparse.Namespace(
            subparser_name='extract',
            input=None,
            sample_bam=[
                os.path.join(CUR_DIR, 'test_data/test_sample1_golden.bam'),
                os.path.join(CUR_DIR, 'test_data/test_sample2_golden.bam')],
            sample_name=['test_sample1', 'test_sample2'],
            sample_type=['tumor', 'tumor'],
            sample_group=['patient1', 'patient1'],
            sample_sex=['M', 'M'],
            database=os.path.join(CUR_DIR, 'test_data/'),
            vcf=os.path.join(CUR_DIR, 'test_data/test.vcf'),
            fafile=os.path.join(CUR_DIR, 'test_data/ref.fasta'),
            bed=os.path.join(CUR_DIR, 'test_data/test.bed'),
            min_mapping_quality=1,
            min_base_quality=1,
            min_coverage=10,
            minor_threshold=0.002,
            major_threshold=0.6,
            discordance_threshold=0.05,
            coverage_threshold=50,
            min_homozygous_thresh=0.1,
            zmin=None,
            zmax=None,
            outdir='.',
            json=None,
            plot=True,
            default_genotype=None,
            overwrite=True,
            no_db_compare=False,
            prefix='test',
            threads=1))
    def setUp(self, mock_args):
        """Set up test fixtures, if any."""

        self.args = get_args()

    def test_load_vcf(self):
        """Test loading the VCF file."""

        extractor = Extract(self.args)

        self.assertGreater(
            len(extractor.sites), 0, msg="Could not parse VCF sites.")
        self.assertEqual(
            len(extractor.sites), 15,
            msg="Did not parse right number of sites.")

    def test_load_bed(self):
        """Test loading the BED file."""

        extractor = Extract(self.args)

        self.assertEqual(
            len(extractor.regions), 1, msg="Expected 1 region in BED file.")

    def test_extract_sample(self):

        extractor = Extract(self.args)
        samples = get_samples(self.args, extraction_mode=True)
        samples = extractor.extract(samples)

        self.assertEqual(len(samples), 2, msg='Did not load 2 samples.')
        self.assertEqual(samples['test_sample1'].sample_name, 'test_sample1', msg='Sample was not loaded correctly.')
        self.assertIsNotNone(samples['test_sample1'].pileup, msg='Sample pileup was not loaded correctly.')
        self.assertEqual(samples['test_sample1'].pileup.shape[0], 15, msg='Did not find pileup for 4 variants. Found: {}.'.format(samples['test_sample1'].pileup))

    def test_sample_minor_contamination(self):
        samples = get_samples(self.args, extraction_mode=False)
        samples = run_minor_contamination(self.args, samples)

        self.assertAlmostEqual(
            samples['test_sample1'].metrics['minor_contamination']['val'], 0.0043,
            places=4, msg='Minor contamination is wrong.')

        self.assertEqual(
            samples['test_sample1'].metrics['minor_contamination']['n_contributing_sites'], 1,
            msg='Count of contributing sites for minor contamination is wrong.')

    def test_sample_major_contamination(self):
        samples = get_samples(self.args, extraction_mode=False)
        samples = run_major_contamination(self.args, samples)

        self.assertAlmostEqual(
            samples['test_sample1'].metrics['major_contamination']['val'], 0.2,
            places=1, msg='Major contamination is wrong.')

    def test_genotyper(self):
        samples = get_samples(self.args, extraction_mode=False)

        genotyper = Genotyper(
            no_db_compare=self.args.no_db_compare,
            discordance_threshold=self.args.discordance_threshold,
            threads=self.args.threads,
            zmin=self.args.zmin,
            zmax=self.args.zmax)
        data = genotyper.compare_samples(samples)

        self.assertEqual(len(data), 4, msg='There were not four comparisons done.')
        self.assertEqual(set(data['Status']), set(['Expected Match']), msg='All sample comparisons were expected to match.')

    def test_sexmismatch(self):
        samples = get_samples(self.args, extraction_mode=False)

        sex_mismatch = SexMismatch(self.args.coverage_threshold)
        results = sex_mismatch.detect_mismatch(samples)

        self.assertEqual(set(results['expected_sex']), set(['M']), msg='Expected all samples to have an expected sex of M.')
        self.assertEqual(set(results['predicted_sex']), set(['M']), msg='Expected all samples to not have a sex mismatch.')



import pandas as pd
import networkx as nx

from biometrics.utils import get_logger

logger = get_logger()


class Cluster:

    def __init__(self, discordance_threshold=0.05):
        self.discordance_threshold = discordance_threshold

    def cluster(self, comparisons):

        assert comparisons is not None, "There is no fingerprint comparison data available."

        if len(comparisons) < 1:
            logger.warning('There are not enough comparisons to cluster.')
            return None

        sample2group = dict(zip(
            comparisons['ReferenceSample'], comparisons['ReferenceSampleGroup']))
        sample2group.update(dict(zip(
            comparisons['QuerySample'], comparisons['QuerySampleGroup'])))

        comparisons['is_same_group'] = comparisons['DiscordanceRate'].map(
            lambda x: 1 if x <= self.discordance_threshold else 0)

        graph = nx.from_pandas_edgelist(
            comparisons[comparisons['is_same_group']==1], 'ReferenceSample', 'QuerySample')

        clusters = []

        for cluster_idx, group in enumerate(nx.connected_components(graph)):
            samples_group = list(group)

            for i, sample in enumerate(samples_group):

                comparisons_sample = comparisons[
                    (comparisons['ReferenceSample']==sample) &
                    (comparisons['QuerySample']!=sample)]
                comparisons_cluster = comparisons_sample[
                    comparisons_sample['QuerySample'].isin(samples_group)]

                sample_status_counts = comparisons_sample['Status'].value_counts()
                cluster_status_counts = comparisons_cluster['Status'].value_counts()

                mean_discordance = 'NA'
                if len(comparisons_cluster) > 0:
                    mean_discordance = comparisons_cluster['DiscordanceRate'].mean()

                row = {
                    'sample_name': sample,
                    'expected_sample_group': sample2group[sample],
                    'cluster_index': cluster_idx,
                    'cluster_size': len(samples_group),
                    'avg_discordance': mean_discordance,
                    'count_expected_matches': cluster_status_counts.get('Expected Match', 0),
                    'count_unexpected_matches': cluster_status_counts.get('Unexpected Match', 0),
                    'count_expected_mismatches': sample_status_counts.get('Expected Mismatch', 0),
                    'count_unexpected_mismatches': sample_status_counts.get('Unexpected Mismatch', 0)
                }
                clusters.append(row)

        clusters = pd.DataFrame(clusters)

        logger.info(
            'Clustering finished. Grouped {} samples into {} clusters. Expected {} clusters.'.format(
            len(sample2group), cluster_idx + 1, len(set(clusters['expected_sample_group']))))

        return clusters



#!/usr/bin/env python
"""Console script for biometrics."""

import sys
import argparse

from biometrics.utils import get_logger
from biometrics.biometrics import run_biometrics

logger = get_logger()


def add_extraction_args(parser):

    parser.add_argument(
        '-i', '--input', action="append", required=False,
        help='''Path to file containing sample information (one per line).
        For example: sample_name,sample_bam,sample_type,sample_sex,sample_group''')
    parser.add_argument(
        '-sb', '--sample-bam', nargs="+", required=False,
        help='''Space-delimited list of BAM files.''')
    parser.add_argument(
        '-st', '--sample-type', nargs="+", required=False,
        help='''Space-delimited list of sample types: Normal or Tumor.
        Must be in the same order as --sample-bam.''')
    parser.add_argument(
        '-ss', '--sample-sex', nargs="+", required=False,
        help='''Space-delimited list of sample sex (i.e. M or F). Must be
        in the same order as --sample-bam.''')
    parser.add_argument(
        '-sg', '--sample-group', nargs="+", required=False,
        help='''Space-delimited list of sample group information
        (e.g. sample patient ID). Must be in the same order as --sample-bam.''')
    parser.add_argument(
        '-sn', '--sample-name', nargs="+", required=False,
        help='''Space-delimited list of sample names. If not specified,
        sample name is automatically figured out from the BAM file. Must
        be in the same order as --sample-bam.''')
    parser.add_argument(
        '--vcf', required=False,
        help='''VCF file containing the sites to be queried.''')
    parser.add_argument(
        '--bed', required=False,
        help='''BED file containing the intervals to be queried.''')
    parser.add_argument(
        '-db', '--database', required=True,
        help='''Directory to store the intermediate files after
        running the extraction step.''')
    parser.add_argument(
        '-ov', '--overwrite', action='store_true',
        help='''Overwrite any existing extraction results.''')
    parser.add_argument(
        '-f', '--fafile', required=True,
        help='''Path to reference fasta file.''')
    parser.add_argument(
        '-q', '--min-mapping-quality', default=1, type=int,
        help='''Minimum mapping quality of reads to be used for pileup.''')
    parser.add_argument(
        '-Q', '--min-base-quality', default=1, type=int,
        help='''Minimum base quality of reads to be used for pileup.''')
    parser.add_argument(
        '-mc', '--min-coverage', default=10, type=int,
        help='''Minimum coverage to count a site.''')
    parser.add_argument(
        '-mht', '--min-homozygous-thresh', default=0.1, type=float,
        help='''Minimum threshold to define homozygous.''')
    parser.add_argument(
        '--default-genotype', default=None,
        help='''Default genotype if coverage is too low (options are Het or Hom).''')
    parser.add_argument(
        '-t', '--threads', default=1, type=int,
        help='''Number of threads to use to extract the samples.''')

    return parser


def add_common_tool_args(parser):
    parser.add_argument(
        '-sn', '--sample-name', nargs="+", required=False,
        help='''Space-delimited list of sample names to analyze.
        Assumes the samples have already been extracted.''')
    parser.add_argument(
        '-i', '--input', action="append", required=False,
        help='''Path to file containing sample information (one per line).
        For example: sample_name,sample_bam,sample_type,sample_sex,sample_group.''')
    parser.add_argument(
        '-db', '--database', required=True,
        help='''Directory to store the intermediate files after
        running the extraction step.''')
    parser.add_argument(
        '-o', '--outdir', default='.',
        help='''Output directory for results.''')
    parser.add_argument(
        '--prefix', default=None,
        help='''Output file prefix.''')
    parser.add_argument(
        '-j', '--json', action='store_true',
        help='''Also output data in JSON format.''')
    parser.add_argument(
        '-nc', '--no-db-compare', action='store_true',
        help='''Do not compare the sample(s) you provided to all samples
        in the database, only compare them with each other.''')

    return parser


def check_arg_equal_len(vals1, vals2, name):

    if vals2 is not None and len(vals1) != len(vals2):
        logger.error(
            '{} does not have the same number of items as --sample-bam'.format(name))
        sys.exit(1)


def check_args(args):

    if args.subparser_name == 'cluster':
        return

    if args.subparser_name != 'extract' and \
            not args.input and not args.sample_name:
        logger.error('You must specify either --input or --sample-name')
        sys.exit(1)

    if args.subparser_name == 'extract' and \
            not args.input and not args.sample_bam:
        logger.error(
            'The extraction tool requires that you specify either --input or --sample-bam')
        sys.exit(1)

    if args.subparser_name == 'extract' and not args.input:
        check_arg_equal_len(args.sample_name, args.sample_bam, '--sample-bam')
        check_arg_equal_len(args.sample_name, args.sample_type, '--sample-type')
        check_arg_equal_len(
            args.sample_name, args.sample_group, '--sample-group')
        check_arg_equal_len(args.sample_name, args.sample_sex, '--sample-sex')


def get_args():

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='''Various tools for fingerprinting samples from BAM files.
        Sample information to each sub command is supplied via input file(s)
        and/or as individual samples.''')
    subparsers = parser.add_subparsers(help='', dest="subparser_name")

    # extract parser

    parser_extract = subparsers.add_parser(
        'extract',
        help='''Intermediate step to extract genotype info from one or more
        samples. The output from this step is required for the rest of the
        fingerprinting tools. However, you do not need to run this step
        manually since it will run automatically if the necessary files
        are missing.''',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_extract = add_extraction_args(parser_extract)

    # sex mismatch parser

    parser_sexmismatch = subparsers.add_parser(
        'sexmismatch', help='Check for sex mismatches.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_sexmismatch = add_common_tool_args(parser_sexmismatch)
    parser_sexmismatch.add_argument(
        '--coverage-threshold', default=50, type=int,
        help='''Samples with Y chromosome above this value will be considered male.''')

    # minor contamination parser

    parser_minor = subparsers.add_parser(
        'minor', help='Check for minor contamination.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_minor = add_common_tool_args(parser_minor)
    parser_minor.add_argument(
        '-p', '--plot', action='store_true',
        help='''Also output plots of the data.''')
    parser_minor.add_argument(
        '--minor-threshold', default=0.002, type=float,
        help='''Minor contamination threshold for bad sample.''')

    # major contamination parser

    parser_major = subparsers.add_parser(
        'major', help='Check for major contamination.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_major = add_common_tool_args(parser_major)
    parser_major.add_argument(
        '-p', '--plot', action='store_true',
        help='''Also output plots of the data.''')
    parser_major.add_argument(
        '--major-threshold', default=0.6, type=float,
        help='''Major contamination threshold for bad sample.''')

    # genotyping parser

    parser_genotype = subparsers.add_parser(
        'genotype', help='Compare sample genotypes to find matches/mismatches.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_genotype = add_common_tool_args(parser_genotype)
    parser_genotype.add_argument(
        '-p', '--plot', action='store_true',
        help='''Also output plots of the data.''')
    parser_genotype.add_argument(
        '--discordance-threshold', default=0.05, type=float,
        help='''Discordance values less than this are regarded
        as matching samples.''')
    parser_genotype.add_argument(
        '-t', '--threads', default=1, type=int,
        help='''Number of threads to use to extract the samples.''')
    parser_genotype.add_argument(
        '--zmin', type=float,
        help='''Minimum z value for the colorscale on the heatmap.''')
    parser_genotype.add_argument(
        '--zmax', type=float,
        help='''Maximum z value for the colorscale on the heatmap.''')

    # cluster parser

    parser_cluster = subparsers.add_parser(
        'cluster', help='Cluster genotype comparison results.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_cluster.add_argument(
        '-i', '--input', action="append", required=True,
        help='''Path to file containing output form \'biometrics genotype\' tool.
        Can be specified more than once.''')
    parser_cluster.add_argument(
        '-o', '--output', default='genotype_clusters.csv',
        help='''Output filename.''')
    parser_cluster.add_argument(
        '--discordance-threshold', default=0.05, type=float,
        help='''Discordance values less than this are regarded
        as matching samples.''')

    args = parser.parse_args()

    check_args(args)

    return args


def main():

    args = get_args()

    run_biometrics(args)

    return 0


if __name__ == "__main__":
    sys.exit(main())



import pandas as pd
import numpy as np


class SexMismatch:
    """
    Class to detect sex mismatch
    """

    def __init__(self, threshold):
        self.threshold = threshold

    def predict_sex(self, sample):

        if sample.region_counts is None:
            return np.nan

        total_count = sample.region_counts['count'].sum()

        predicted_sex = 'M' if total_count > self.threshold else 'F'

        return predicted_sex

    def detect_mismatch(self, samples):

        results = []

        for i, sample_name in enumerate(samples):

            sample = samples[sample_name]

            predicted_sex = self.predict_sex(sample)

            results.append({
                'sample': sample_name,
                'expected_sex': sample.sample_sex,
                'predicted_sex': predicted_sex
            })

        results = pd.DataFrame(results)

        results['sex_mismatch'] = \
            (results['expected_sex'] != results['predicted_sex']).astype(str)
        results.loc[pd.isna(results['predicted_sex']), 'sex_mismatch'] = np.nan

        return results



"""Top-level package for biometrics."""

import os

__author__ = """Charlie Murphy"""
__email__ = 'murphyc4@mskcc.org'

# version info

library_path = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(library_path, "VERSION"), "r") as fh:
    __version__ = fh.read().strip()



import sys
import os
from multiprocessing import Pool

import pandas as pd
import numpy as np
import plotly.graph_objects as go

from biometrics.utils import get_logger

EPSILON = 1e-9
logger = get_logger()


class Genotyper:

    def __init__(self, no_db_compare, discordance_threshold=0.05, threads=1, zmin=None, zmax=None):
        self.no_db_compare = no_db_compare
        self.discordance_threshold = discordance_threshold
        self.threads = threads
        self.zmax = zmax
        self.zmin = zmin
        self.sample_type_ratio = 1
        self.comparisons = None

    def are_samples_same_group(self, sample1, sample2):

        if sample1.sample_group is None or sample2.sample_group is None:
            return np.nan

        if sample1.sample_group == sample2.sample_group:
            return True
        else:
            return False

    def _plot_heatmap(self, data, outdir, name, title="Discordance calculations between samples", size_ratio=None):

        width = None
        height = None

        if size_ratio is not None and size_ratio != 1:
            width = 1400
            height = (width * size_ratio)/4

        fig = go.Figure()
        fig.add_trace(
            go.Heatmap(
                x=data['ReferenceSample'],
                y=data['QuerySample'],
                z=data['DiscordanceRate'],
                legendgroup="Discordance",
                name='Discordance',
                customdata=data.to_numpy(),
                hovertemplate='<b>Reference sample:</b> %{customdata[0]}' +
                              '<br><b>Query sample:</b> %{customdata[1]}' +
                              '<br><b>Homozygous count in reference:</b> %{customdata[3]}' +
                              '<br><b>Total match count:</b> %{customdata[4]}' +
                              '<br><b>Homozygous match count:</b> %{customdata[5]}' +
                              '<br><b>Heterozygous match count:</b> %{customdata[6]}' +
                              '<br><b>Homozygous mismatch count:</b> %{customdata[7]}' +
                              '<br><b>Heterozygous mismatch count:</b> %{customdata[8]}' +
                              '<br><b>Discordance rate:</b> %{customdata[9]}' +
                              '<br><b>Status:</b> %{customdata[12]}' +
                              '<extra></extra>',
                zmin=self.zmin,
                zmax=self.zmax,
                colorscale='Blues_r'
            ))

        # add red dots to sample pairs that are unexpected match/mismatch

        data_sub = data[(data['Status']=='Unexpected Match') | (data['Status']=='Unexpected Mismatch')].copy()

        if len(data_sub) > 0:
            fig.add_trace(
                go.Scatter(
                    mode="markers",
                    x=data_sub['ReferenceSample'],
                    y=data_sub['QuerySample'],
                    marker_symbol=[17],
                    marker_color="red",
                    marker_line_width=0,
                    marker_size=10,
                    customdata=data_sub.to_numpy(),
                    hovertemplate='%{customdata[12]}<extra></extra>'))

        fig.update_layout(
            yaxis_title="Query samples",
            xaxis_title="Reference samples",
            legend_title_text="Discordance",
            title_text=title,
            width=width, height=height)
        fig.write_html(os.path.join(outdir, name))

    def plot(self, data, outdir):

        # make plot for comparing input samples with each other

        data_sub = data[~data['DatabaseComparison']].copy()
        data_sub['DiscordanceRate'] = data_sub['DiscordanceRate'].map(
            lambda x: round(x, 4))

        if data_sub.shape[0] > 1:
            self._plot_heatmap(
                data_sub, outdir, name='genotype_comparison_input_only.html',
                title="Discordance calculations between input samples")

        # make plot for comparing input samples with database samples

        data_sub = data[data['DatabaseComparison']].copy()
        data_sub['DiscordanceRate'] = data_sub['DiscordanceRate'].map(
            lambda x: round(x, 4))

        if data_sub.shape[0] > 1:
            self._plot_heatmap(
                data_sub, outdir, name='genotype_comparison_database.html',
                title="Discordance calculations between input samples and database samples",
                size_ratio=self.sample_type_ratio)

    def _compute_discordance(self, reference_sample, query_sample):
        """
        Compute discordance between two samples
        """

        pileup_ref = reference_sample.pileup
        pileup_query = query_sample.pileup

        common_covered = ~pd.isna(pileup_ref['genotype_class']) & ~pd.isna(pileup_query['genotype_class'])
        pileup_ref = pileup_ref[common_covered]
        pileup_query = pileup_query[common_covered]

        row = {
            'ReferenceSample': reference_sample.sample_name,
            'ReferenceSampleGroup': reference_sample.sample_group,
            'QuerySample': query_sample.sample_name,
            'QuerySampleGroup': query_sample.sample_group}

        if len(pileup_query) > 0:
            row['HomozygousInRef'] = sum(pileup_ref['genotype_class'] == 'Hom')
            row['TotalMatch'] = sum(pileup_ref['genotype_class'] == pileup_query['genotype_class'])
            row['HomozygousMatch'] = sum((pileup_ref['genotype_class'] == pileup_query['genotype_class']) & (pileup_ref['genotype_class'] == 'Hom'))
            row['HeterozygousMatch'] = sum((pileup_ref['genotype_class'] == pileup_query['genotype_class']) & (pileup_ref['genotype_class'] == 'Het'))
            row['HomozygousMismatch'] = sum((pileup_ref['genotype'] != pileup_query['genotype']) & ((pileup_ref['genotype_class'] == 'Hom') & (pileup_query['genotype_class'] == 'Hom')))
            row['HeterozygousMismatch'] = sum((pileup_ref['genotype_class'] != pileup_query['genotype_class']) & ((pileup_ref['genotype_class'] == 'Het') | (pileup_query['genotype_class'] == 'Het')))
            row['CountOfCommonSites'] = len(pileup_query)
        else:
            # if there are no regions with enough coverage

            row['HomozygousInRef'] = np.nan
            row['TotalMatch'] = np.nan
            row['HomozygousMatch'] = np.nan
            row['HeterozygousMatch'] = np.nan
            row['HomozygousMismatch'] = np.nan
            row['HeterozygousMismatch'] = np.nan
            row['CountOfCommonSites'] = 0

        return row

    def _compute_discordance_batch_job(self, batch_sample_pairs):
        """
        batch job to take a batch of pairs of samples to compute discordance
        """
        rows = []

        for reference_sample, query_sample in batch_sample_pairs:
            rows.append(self._compute_discordance(
                reference_sample, query_sample))

        return rows

    def _compare_sample_lists(self, sample_set1, sample_set2, samples):
        """
        Compare two lists of samples. Does so by first grouping the sample
        comparisons into batches for parallel processing
        """

        jobs = []
        thread_pool = Pool(self.threads)
        total_comparisons = len(sample_set1) * len(sample_set2)
        parallel_batch_size = max(int(total_comparisons / self.threads), 1)

        current_batch = []

        for i, sample_name1 in enumerate(sample_set1):
            for j, sample_name2 in enumerate(sample_set2):
                current_batch.append([samples[sample_name1], samples[sample_name2]])

                if len(current_batch) >= parallel_batch_size:
                    jobs.append(current_batch)
                    current_batch = []

        # add any remaining jobs

        if len(current_batch) > 0:
            jobs.append(current_batch)

        # analyze, collect results, and flatten the lists

        results = thread_pool.map(self._compute_discordance_batch_job, jobs)
        results = [item for sublist in results for item in sublist]

        return results


    def compare_samples(self, samples):

        comparisons = []
        samples_db = dict(filter(lambda x: x[1].query_group, samples.items()))
        samples_input = dict(filter(
            lambda x: not x[1].query_group, samples.items()))

        # get the number of each type of sample and compute a ratio
        # this is used to plot the heatmap when comparing with database
        # samples

        sample_n_db = len(samples_db)
        sample_n_input = len(samples_input)

        if sample_n_db > 0 and sample_n_input > 0 and sample_n_db > sample_n_input:
            self.sample_type_ratio = sample_n_db / sample_n_input

        # check to see if there are appropriate number of samples to
        # do the analysis

        if self.no_db_compare:
            if len(samples_input) <= 1:
                logger.error("You need to specify 2 or more samples in order to compare genotypes.")
                sys.exit(1)
        else:
            if len(samples_input) <= 1 and len(samples_db) < 1:
                logger.error("There are no samples in the database to compare with")
                sys.exit(1)

        # compare all the input samples to each other

        if sample_n_input > 1:
            results = self._compare_sample_lists(
                samples_input, samples_input, samples)

            for i in range(len(results)):
                results[i]['DatabaseComparison'] = True
            comparisons += results

        # for each input sample, compare with all the samples in the db

        if not self.no_db_compare and sample_n_db > 0:
            results = self._compare_sample_lists(
                samples_input, samples_db, samples)

            for i in range(len(results)):
                results[i]['DatabaseComparison'] = True
            comparisons += results

        comparisons = pd.DataFrame(comparisons)

        # compute discordance rate

        comparisons['DiscordanceRate'] = comparisons['HomozygousMismatch'] / (comparisons['HomozygousInRef'] + EPSILON)
        # data['DiscordanceRate'] = data['DiscordanceRate'].map(lambda x: round(x, 6))
        comparisons.loc[comparisons['HomozygousInRef'] < 10, 'DiscordanceRate'] = np.nan

        # for each comparison, indicate if the match/mismatch is expected
        # or not expected

        comparisons['Matched'] = comparisons['DiscordanceRate'] < self.discordance_threshold
        comparisons['ExpectedMatch'] = comparisons.apply(
            lambda x: self.are_samples_same_group(
                samples[x['ReferenceSample']],
                samples[x['QuerySample']]), axis=1)

        comparisons['Status'] = ''
        comparisons.loc[comparisons.Matched & comparisons.ExpectedMatch, 'Status'] = "Expected Match"
        comparisons.loc[comparisons.Matched & ~comparisons.ExpectedMatch, 'Status'] = "Unexpected Match"
        comparisons.loc[~comparisons.Matched & comparisons.ExpectedMatch, 'Status'] = "Unexpected Mismatch"
        comparisons.loc[~comparisons.Matched & ~comparisons.ExpectedMatch, 'Status'] = "Expected Mismatch"

        self.comparisons = comparisons[[
            'ReferenceSample', 'ReferenceSampleGroup', 'QuerySample', 'QuerySampleGroup',
            'CountOfCommonSites', 'HomozygousInRef', 'TotalMatch', 'HomozygousMatch',
            'HeterozygousMatch', 'HomozygousMismatch',
            'HeterozygousMismatch', 'DiscordanceRate', 'Matched',
            'ExpectedMatch', 'Status']]

        logger.info('Total comparisons: {}'.format(len(comparisons)))
        logger.info('Count of expected matches: {}'.format(
            len(comparisons[comparisons['Status']=="Expected Match"])))
        logger.info('Count of unexpected matches: {}'.format(
            len(comparisons[comparisons['Status']=="Unexpected Match"])))
        logger.info('Count of unexpected mismatches: {}'.format(
            len(comparisons[comparisons['Status']=="Unexpected Mismatch"])))
        logger.info('Count of expected mismatches: {}'.format(
            len(comparisons[comparisons['Status']=="Expected Mismatch"])))

        return self.comparisons



import os

import pandas as pd
import numpy as np
import plotly.graph_objects as go


class MinorContamination():
    """
    Minor contamination.
    """

    def __init__(self, threshold):
        self.threshold = threshold

    def to_dataframe(self, samples):

        data = pd.DataFrame(
            columns=['sample_name', 'sample_group', 'sample_sex', 'sample_type',
                     'total_homozygous_sites', 'n_contributing_sites', 'minor_contamination'])

        for sample_name, sample in samples.items():

            row = {
                'sample_name': sample.sample_name,
                'sample_group': sample.sample_group,
                'sample_sex': sample.sample_sex,
                'sample_type': sample.sample_type,
                'total_homozygous_sites': sample.metrics['minor_contamination']['n_homozygous_sites'],
                'n_contributing_sites': sample.metrics['minor_contamination']['n_contributing_sites'],
                'minor_contamination': sample.metrics['minor_contamination']['val']
            }

            data = data.append(row, ignore_index=True)

        data = data.sort_values('minor_contamination', ascending=False)
        return data

    def plot(self, samples, outdir):
        """
        Plot major contamination data.
        """

        data = self.to_dataframe(samples)
        data['minor_contamination'] = data['minor_contamination'].map(
            lambda x: round(x, 5))

        fig = go.Figure()
        fig.add_trace(
            go.Bar(
                x=data['sample_name'],
                y=data['minor_contamination'],
                customdata=data.to_numpy(),
                hovertemplate='<b>Sample group:</b> %{customdata[1]}' +
                              '<br><b>Sample name:</b> %{customdata[0]}' +
                              '<br><b>Sample sex:</b> %{customdata[2]}' +
                              '<br><b>Sample type:</b> %{customdata[3]}' +
                              '<br><b>Total homozygous sites:</b> %{customdata[4]}' +
                              '<br><b>Total contributing sites:</b> %{customdata[5]}' +
                              '<br><b>Minor contamination:</b> %{y:E}' +
                              '<extra></extra>'))

        fig.update_layout(
            yaxis_title="Minor contamination",
            title_text="Minor contamination across samples")
        fig.add_hline(y=self.threshold, line_color='red')

        fig.write_html(os.path.join(outdir, 'minor_contamination.html'))

        # plot VAF of contributing sites

        plot_data = []

        for i, sample_name in enumerate(data['sample_name']):
            contributing_sites = samples[sample_name].metrics['minor_contamination']['contributing_sites']

            for site_id, site_data in contributing_sites.items():
                site_data['sample_name'] = sample_name
                site_data['MAF'] = site_data['minor_allele_freq']
                site_data['index'] = i

                del site_data['minor_allele_freq']
                plot_data.append(site_data)

        plot_data = pd.DataFrame(plot_data)
        plot_data = plot_data[[
            'sample_name', 'chrom', 'pos', 'ref', 'alt', 'MAF', 'reads_all', 'A', 'C', 'T', 'G',
            'N', 'index']]
        plot_data['MAF'] = plot_data['MAF'].map(lambda x: round(x, 5))


        fig = go.Figure()

        fig.add_trace(
            go.Scatter(
                x=plot_data['index'],
                y=plot_data['MAF'],
                mode='markers',
                showlegend=False,
                customdata=plot_data.to_numpy(),
                hovertemplate='<b>Sample:</b> %{customdata[0]}' +
                              '<br><b>Chrom:</b> %{customdata[1]}' +
                              '<br><b>Pos:</b> %{customdata[2]}' +
                              '<br><b>Ref allele:</b> %{customdata[3]}' +
                              '<br><b>Alt allele:</b> %{customdata[4]}' +
                              '<br><b>MAF:</b> %{customdata[5]}' +
                              '<br><b>Total reads:</b> %{customdata[6]}' +
                              '<br><b>Count A:</b> %{customdata[7]}' +
                              '<br><b>Count C:</b> %{customdata[8]}' +
                              '<br><b>Count T:</b> %{customdata[9]}' +
                              '<br><b>Count G:</b> %{customdata[10]}' +
                              '<br><b>Count N:</b> %{customdata[11]}' +
                              '<extra></extra>'))

        data = data[data['sample_name'].isin(plot_data['sample_name'])]
        data['index'] = range(len(data))

        for i in data.index:
            fig.add_shape(go.layout.Shape(
                type="line",
                x0=data.at[i, 'index'] - 0.5,
                y0=data.at[i, 'minor_contamination'],
                x1=data.at[i, 'index'] + 0.5,
                y1=data.at[i, 'minor_contamination'],
                line=dict(color='black', width=2)
            ))

        fig.add_trace(go.Scatter(
            x=[1],
            y=[3],
            name='Minor contamination',
            line_color='black'
        ))
        fig.add_trace(go.Scatter(
            x=[1],
            y=[3],
            name='Threshold',
            line_color='red'
        ))
        fig.add_trace(go.Scatter(
            x=[1],
            y=[3],
            name='Minor allele sites (MAF > 0)',
            line_color='#636EFA',
            mode='markers'
        ))


        ticks = plot_data[['sample_name', 'index']].drop_duplicates()

        fig.update_layout(
            yaxis_title="Minor allele frequency",
            title_text="Statistics of sites that contribute to minor contamination",
            yaxis = dict(range=(-0.005, plot_data['MAF'].max()*1.05)),
            xaxis = dict(
                tickmode = 'array',
                tickvals = ticks['index'],
                ticktext = ticks['sample_name']
            ))
        fig.add_hline(y=self.threshold, line_color='red')

        fig.write_html(os.path.join(outdir, 'minor_contamination_sites.html'))

        return data

    def estimate(self, samples):
        """
        Estimate minor contamination.
        """

        for sample_name, sample in samples.items():

            sites = sample.pileup
            sites_notna = sites[~pd.isna(sites['genotype_class'])]
            hom_sites = sites_notna[sites_notna['genotype_class'] == 'Hom']

            sample.metrics['minor_contamination'] = {}
            sample.metrics['minor_contamination']['n_homozygous_sites'] = len(hom_sites)

            if len(hom_sites) == 0:
                sample.metrics['minor_contamination']['val'] = np.nan
                sample.metrics['minor_contamination']['n_contributing_sites'] = np.nan
                sample.metrics['minor_contamination']['contributing_sites'] = np.nan
            else:

                contributing_sites = hom_sites[hom_sites['minor_allele_freq']>0]
                contributing_sites.index = contributing_sites['chrom'].astype(str) + ':' + \
                    contributing_sites['pos'].astype(str)

                sample.metrics['minor_contamination']['val'] = \
                    hom_sites['minor_allele_freq'].mean()
                sample.metrics['minor_contamination']['n_contributing_sites'] = \
                    len(contributing_sites)
                sample.metrics['minor_contamination']['contributing_sites'] = \
                    contributing_sites.to_dict(orient='index')


        return samples



import os
from multiprocessing import Pool

import pandas as pd
import numpy as np
import vcf
from pysam import AlignmentFile


class Extract:
    """
    Class for extracting genotype information from alignment file using
    the user supplied VCF file.
    """

    def __init__(self, args):
        self.db = args.database
        self.threads = args.threads
        self.min_mapping_quality = args.min_mapping_quality
        self.min_base_quality = args.min_base_quality
        self.default_genotype = args.default_genotype
        self.vcf = args.vcf
        self.bed = args.bed
        self.fafile = args.fafile
        self.overwrite = args.overwrite
        self.min_coverage = args.min_coverage
        self.min_homozygous_thresh = args.min_homozygous_thresh
        self.sites = []
        self.regions = None

        self._parse_vcf()
        self._parse_bed_file()

    def _parse_vcf(self):

        if self.vcf is None:
            return

        for record in vcf.Reader(open(self.vcf, 'r')):
            self.sites.append({
                'chrom': record.CHROM,
                'start': record.POS-1,
                'end': record.POS,
                'ref_allele': str(record.REF),
                'alt_allele': str(record.ALT[0])
            })

    def _parse_bed_file(self):

        if self.bed is None:
            return

        self.regions = pd.read_csv(self.bed, sep='\t', header=None)
        self.regions.columns = range(self.regions.shape[1])

    def _extract_regions(self, sample):
        """
        Code to extract the coverage information for the regions listed
        in the BED file.
        """

        if self.regions is None:
            return sample

        # get the pileup

        bam = AlignmentFile(sample.sample_bam)
        region_counts = []

        for i in self.regions.index:

            chrom = self.regions.at[i, 0]
            start = int(self.regions.at[i, 1])
            end = int(self.regions.at[i, 2])

            count = bam.count(chrom, start, end)

            region_counts.append({
                'chrom': chrom,
                'start': start,
                'end': end,
                'count': count})

        region_counts = pd.DataFrame(region_counts)

        sample.region_counts = region_counts

        return sample

    def _get_minor_allele_freq(self, allele_counts):

        coverage = sum(allele_counts)

        if coverage < self.min_coverage or coverage == 0:
            return np.nan
        else:
            return min(allele_counts) / coverage

    def _get_genotype_class(self, minor_allele_freq):
        """
        Determine if Het, Hom, or unknown/NA.
        """

        if pd.isna(minor_allele_freq):

            if self.default_genotype is not None:
                return self.default_genotype

            return np.nan
        else:
            if minor_allele_freq <= self.min_homozygous_thresh:
                return 'Hom'
            else:
                return 'Het'

    def _get_genotype(self, genotype, allele_counts, alleles):
        """
        Get the genotype in terms of the allele(s) (e.g. A, T, AT, GC, etc.)
        """

        if pd.isna(genotype):
            return np.nan
        elif genotype == 'Het':
            return ''.join(alleles)
        else:
            if allele_counts[0] > allele_counts[1]:
                return alleles[0]
            else:
                return alleles[1]

    def _get_genotype_info(self, pileup_site, ref_allele, alt_allele):
        """
        Plot minor contamination data.
        """

        allele_counts = [pileup_site[ref_allele], pileup_site[alt_allele]]

        pileup_site['minor_allele_freq'] = self._get_minor_allele_freq(
            allele_counts)

        pileup_site['genotype_class'] = self._get_genotype_class(
            pileup_site['minor_allele_freq'])

        pileup_site['genotype'] = self._get_genotype(
            pileup_site['genotype_class'], allele_counts,
            [ref_allele, alt_allele])

        return pileup_site

    def _add_base(self, site, old_base, old_base_qual, new_base,
                  new_base_qual):
        """
        This function is for dealing with the various scenarios that can
        arise when a read pair overlaps and how to handle when the
        bases mismatch. The 'old_base' refers to the first base observed when
        computing pileup information (usually the forward read). Then the
        'new_base' is from the second read in the overlaping pair.
        """

        if old_base is None:
            return [new_base, new_base_qual]

        if old_base == new_base:
            return [old_base, old_base_qual]

        if old_base != 'N' and new_base != 'N':
            if new_base == site['ref_allele']:
                return [new_base, new_base_qual]
            else:
                return [old_base, old_base_qual]

        if old_base == site['ref_allele']:
            return [old_base, old_base_qual]
        elif new_base == site['ref_allele']:
            return [new_base, new_base_qual]
        elif old_base == site['alt_allele'] and old_base_qual >= '!':
            return [old_base, old_base_qual]
        elif new_base == site['alt_allele'] and new_base_qual >= '!':
            return [new_base, new_base_qual]
        else:
            return ['N', '&']

    def _pileup(self, bam, site):
        """
        Get the per-site pileup information.
        """

        read_data = {}
        allele_counts = {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'N': 0}

        for pileupcolumn in bam.pileup(
                contig=site['chrom'], start=site['start'], end=site['end'],
                truncate=True, max_depth=30000, stepper='nofilter',
                min_base_quality=self.min_base_quality):

            for pileupread in pileupcolumn.pileups:

                if pileupread.query_position is None:
                    continue

                mapq = pileupread.alignment.mapping_quality
                read_name = pileupread.alignment.qname
                base_qual = pileupread.alignment.qual[pileupread.query_position]
                base = pileupread.alignment.query_sequence[pileupread.query_position]

                if (mapq < self.min_mapping_quality) or pileupread.is_refskip or pileupread.is_del:
                    # skip the read if its mapping quality is too low
                    # or if the site is part of an indel
                    continue

                if read_name in read_data and read_data[read_name][0] == 'N':
                    continue
                elif read_name in read_data:
                    vals = self._add_base(
                        site, read_data[read_name][0],
                        read_data[read_name][1], base, base_qual)
                    read_data[read_name] = vals[0:2]
                else:
                    read_data[read_name] = [base, base_qual]

        total = 0
        matches = 0
        mismatches = 0
        for read, base_data in read_data.items():

            allele_counts[base_data[0]] += 1
            total += 1

            if base_data[0] == site['ref_allele']:
                matches += 1
            else:
                mismatches += 1

        return {
            'chrom': site['chrom'],
            'pos': site['end'],
            'ref': site['ref_allele'],
            'alt': site['alt_allele'],
            'reads_all': total,
            'matches': matches,
            'mismatches': mismatches,
            'A': allele_counts['A'],
            'C': allele_counts['C'],
            'T': allele_counts['T'],
            'G': allele_counts['G'],
            'N': allele_counts['N']
        }

    def _extract_sites(self, sample):
        """
        Loop through all positions and get pileup information.
        """

        if not self.sites:
            return sample

        # get the pileup

        bam = AlignmentFile(sample.sample_bam)
        pileup = pd.DataFrame()

        for site in self.sites:

            pileup_site = self._pileup(bam, site)

            pileup_site = self._get_genotype_info(
                pileup_site, site['ref_allele'], site['alt_allele'])

            pileup = pileup.append(pileup_site, ignore_index=True)

        pileup = pileup[[
            'chrom', 'pos', 'ref', 'alt', 'reads_all', 'matches', 'mismatches',
            'A', 'C', 'T', 'G', 'N', 'minor_allele_freq', 'genotype_class',
            'genotype']]

        for col in ['pos', 'A', 'C', 'T', 'G', 'N', 'matches', 'mismatches', 'reads_all']:
            pileup[col] = pileup[col].astype(int)

        sample.pileup = pileup

        return sample

    def _extraction_job(self, sample):
        """
        Function to do the extraction steps for a single sample.
        Supposed to be called by multiprocessing functions to parallelize it.
        """

        sample = self._extract_sites(sample)
        sample = self._extract_regions(sample)
        sample.save_to_file()

        return sample

    def extract(self, samples):
        """
        Function to call to extract the pileup and region information
        for the given samples.
        """

        if type(samples) != dict:
            samples = {samples.sample_name: samples}

        # determine with samples need to be extracted, and put them in a list

        samples_to_extract = []

        for sample_name, sample in samples.items():

            # if extraction file exists then load it

            if os.path.exists(sample.extraction_file) and not self.overwrite:
                sample.load_from_file()
                continue

            samples_to_extract.append(sample)

        # if any samples need to be extracted, then do so
        # (using multiprocessing)

        if len(samples_to_extract) > 0:

            thread_pool = Pool(self.threads)

            samples_processed = thread_pool.map(
                self._extraction_job, samples_to_extract)

            for sample in samples_processed:
                samples[sample.sample_name] = sample

        return samples



import os
import glob

import pandas as pd

from biometrics.sample import Sample
from biometrics.extract import Extract
from biometrics.genotype import Genotyper
from biometrics.cluster import Cluster
from biometrics.minor_contamination import MinorContamination
from biometrics.major_contamination import MajorContamination
from biometrics.sex_mismatch import SexMismatch
from biometrics.utils import standardize_sex_nomenclature, get_logger

logger = get_logger()


def write_to_file(args, data, basename):
    """
    Generic function to save output to a file.
    """

    outdir = os.path.abspath(args.outdir)

    outpath = os.path.join(outdir, basename + '.csv')
    data.to_csv(outpath, index=False)

    if args.json:
        outpath = os.path.join(outdir, basename + '.json')
        data.to_json(outpath)


def run_extract(args, samples):
    """
    Extract the pileup and region information from the samples. Then
    save to the database.
    """

    extractor = Extract(args=args)
    samples = extractor.extract(samples)

    return samples


def run_sexmismatch(args, samples):
    """
    Find and sex mismatches and save the output
    """

    sex_mismatch = SexMismatch(args.coverage_threshold)

    results = sex_mismatch.detect_mismatch(samples)

    basename = 'sex_mismatch'
    if args.prefix:
        basename = args.prefix + '_' + basename

    write_to_file(args, results, basename)


def run_minor_contamination(args, samples):
    """
    Compute minor contamination and save the output and figure
    """

    minor_contamination = MinorContamination(threshold=args.minor_threshold)
    samples = minor_contamination.estimate(samples)

    data = minor_contamination.to_dataframe(samples)

    basename = 'minor_contamination'
    if args.prefix:
        basename = args.prefix + '_' + basename

    write_to_file(args, data, basename)

    if args.plot:
        if len(samples) > 1000:
            logger.warning('Turning off plotting functionality. You are trying to plot more than 1000 samples, which is too cumbersome.')
        else:
            minor_contamination.plot(samples, args.outdir)

    return samples


def run_major_contamination(args, samples):
    """
    Compute major contamination and save the output and figure.
    """

    major_contamination = MajorContamination(threshold=args.major_threshold)
    samples = major_contamination.estimate(samples)

    data = major_contamination.to_dataframe(samples)

    basename = 'major_contamination'
    if args.prefix:
        basename = args.prefix + '_' + basename

    write_to_file(args, data, basename)

    if args.plot:
        if len(samples) > 1000:
            logger.warning('Turning off plotting functionality. You are trying to plot more than 1000 samples, which is too cumbersome.')
        else:
            major_contamination.plot(samples, args.outdir)

    return samples


def run_genotyping(args, samples):
    """
    Run the genotyper and save the output and figure.
    """

    genotyper = Genotyper(
        no_db_compare=args.no_db_compare,
        discordance_threshold=args.discordance_threshold,
        threads=args.threads,
        zmin=args.zmin,
        zmax=args.zmax)
    cluster_handler = Cluster(args.discordance_threshold)
    comparisons = genotyper.compare_samples(samples)

    # save genotyping output

    basename = 'genotype_comparison'
    if args.prefix:
        basename = args.prefix + '_' + basename

    write_to_file(args, comparisons, basename)

    # cluster just the input samples

    samples_input = dict(filter(
        lambda x: not x[1].query_group, samples.items()))

    samples_names = [i.sample_name for i in samples_input.values()]
    comparisons_input = comparisons[
        (comparisons['ReferenceSample'].isin(samples_names)) &
        (comparisons['QuerySample'].isin(samples_names))].copy()

    logger.info('Clustering input samples...')
    clusters = cluster_handler.cluster(comparisons_input)

    if clusters is not None:
        basename = 'genotype_clusters_input'
        if args.prefix:
            basename = args.prefix + '_' + basename

        write_to_file(args, clusters, basename)

    # cluster all the samples

    if not args.no_db_compare:

        are_there_db_samples = len(samples_input) != len(samples)

        if not are_there_db_samples:
            logger.warning(
                'The set of database and input samples are the same. Will only cluster the samples once.')
        else:
            logger.info('Clustering input and database samples...')
            clusters = cluster_handler.cluster(comparisons)

            if clusters is not None:
                basename = 'genotype_clusters_database'
                if args.prefix:
                    basename = args.prefix + '_' + basename

                write_to_file(args, clusters, basename)

    # save plots

    if args.plot:
        if len(samples) > 1000:
            logger.warning('Turning off plotting functionality. You are trying to plot more than 1000 samples, which is too cumbersome.')
        else:
            genotyper.plot(comparisons, args.outdir)

    return samples


def run_cluster(args):

    comparisons = []
    for input in args.input:
        comparisons.append(
            pd.read_csv(input)
        )
    comparisons = pd.concat(comparisons)
    comparisons = comparisons.drop_duplicates(['ReferenceSample', 'QuerySample'])

    cluster_handler = Cluster(args.discordance_threshold)

    logger.info('Clustering input samples...')
    clusters = cluster_handler.cluster(comparisons)

    if clusters is not None:
        clusters.to_csv(args.output, index=False)


def load_input_sample_from_db(sample_name, database):
    """
    Loads any the given (that the user specified via the CLI) from the
    database.
    """

    extraction_file = os.path.join(database, sample_name + '.pk')

    assert os.path.exists(extraction_file), 'Could not find: {}. Please rerun the extraction step.'.format(
        extraction_file)

    sample = Sample(query_group=False)
    sample.load_from_file(extraction_file)

    return sample


def load_database_samples(database, existing_samples):
    """
    Loads any samples that are already present in the database AND
    which were not specified as input via the CLI.
    """

    samples = {}

    for pickle_file in glob.glob(os.path.join(database, '*pk')):

        sample_name = os.path.basename(pickle_file).replace('.pk', '')

        if sample_name in existing_samples:
            continue

        sample = Sample(db=database, query_group=True)
        sample.load_from_file(extraction_file=pickle_file)

        samples[sample.sample_name] = sample

    return samples


def get_samples_from_input(input, database, extraction_mode):
    """
    Parse the sample information from the user-supplied CSV file.
    """

    samples = {}

    for fpath in input:

        input = pd.read_csv(fpath, sep=',')

        # check if some required columns are present

        assert 'sample_bam' in input.columns, 'Input file does not have the \'sample_bam\' column.'
        assert 'sample_name' in input.columns, 'Input does not have \'sample_name\' column.'

        input = input.to_dict(orient='records')

        for row in input:

            if not extraction_mode:
                # if not running extract tool, then just need to get
                # the sample name

                sample_name = row['sample_name']

                sample = load_input_sample_from_db(sample_name, database)
                samples[sample.sample_name] = sample

                continue

            # parse in the input

            sample = Sample(
                sample_name=row['sample_name'],
                sample_bam=row['sample_bam'],
                sample_group=row.get('sample_group'),
                sample_type=row.get('sample_type'),
                sample_sex=standardize_sex_nomenclature(row.get('sample_sex')),
                db=database)

            samples[sample.sample_name] = sample

    return samples


def get_samples_from_bam(args):
    """
    Parse the sample information the user supplied via the CLI.
    """

    samples = {}

    for i, sample_bam in enumerate(args.sample_bam):

        sample_sex = standardize_sex_nomenclature(
            args.sample_sex[i] if args.sample_sex is not None else None)
        sample_name = args.sample_name[i] if args.sample_name is not None else None
        sample_group = args.sample_group[i] \
            if args.sample_group is not None else None
        sample_type = args.sample_type[i] \
            if args.sample_type is not None else None

        sample = Sample(
            sample_bam=sample_bam, sample_group=sample_group,
            sample_name=sample_name, sample_type=sample_type,
            sample_sex=sample_sex, db=args.database)

        samples[sample.sample_name] = sample

    return samples


def get_samples_from_name(sample_names, database):
    """
    Parse the sample information the user supplied via the CLI.
    """

    samples = {}

    for i, sample_name in enumerate(sample_names):
        sample = load_input_sample_from_db(sample_name, database)
        samples[sample.sample_name] = sample

    return samples


def get_samples(args, extraction_mode=False):
    """
    Parse the sample information the user supplied via the CLI.
    """

    samples = {}

    if args.input:
        samples.update(get_samples_from_input(
            args.input, args.database, extraction_mode))

    if extraction_mode:
        if args.sample_bam:
            samples.update(get_samples_from_bam(args))
    else:
        if args.sample_name:
            samples.update(get_samples_from_name(
                args.sample_name, args.database))

        for sample_name in samples.keys():
            extration_file = os.path.join(args.database, sample_name + '.pk')
            samples[sample_name].load_from_file(extration_file)

        existing_samples = set([i for i in samples.keys()])

        if not args.no_db_compare:
            samples.update(load_database_samples(
                args.database, existing_samples))

    return samples


def create_outdir(outdir):
    os.makedirs(outdir, exist_ok=True)


def run_biometrics(args):
    """
    Decide what tool to run based in CLI input.
    """

    if args.subparser_name == 'cluster':
        run_cluster(args)
        return

    extraction_mode = args.subparser_name == 'extract'

    samples = get_samples(args, extraction_mode=extraction_mode)

    # if not extraction_mode and args.plot:

    if extraction_mode:
        create_outdir(args.database)
        run_extract(args, samples)
    elif args.subparser_name == 'sexmismatch':
        create_outdir(args.outdir)
        run_sexmismatch(args, samples)
    elif args.subparser_name == 'minor':
        create_outdir(args.outdir)
        run_minor_contamination(args, samples)
    elif args.subparser_name == 'major':
        create_outdir(args.outdir)
        run_major_contamination(args, samples)
    elif args.subparser_name == 'genotype':
        create_outdir(args.outdir)
        run_genotyping(args, samples)



import pickle
import os

import pandas as pd


class Sample:
    """
    Class to hold information related to a single sample.
    """

    def __init__(self, sample_name=None, sample_bam=None, sample_group=None,
                 sample_sex=None, sample_type=None, db=None, query_group=False):
        self.sample_bam = sample_bam
        self.sample_name = sample_name
        self.sample_sex = sample_sex
        self.sample_group = sample_group
        self.sample_type = sample_type

        self.pileup = None
        self.region_counts = None
        self.extraction_file = None
        self.query_group = query_group
        self.metrics = {}

        if db is not None and self.sample_name is not None:
            self.extraction_file = os.path.join(db, self.sample_name + '.pk')

    def save_to_file(self):

        pileup_data = self.pileup.to_dict("records")

        if self.region_counts is not None:
            region_counts = self.region_counts.to_dict('records')
        else:
            region_counts = None

        sample_data = {
            'sample_bam': self.sample_bam,
            'sample_name': self.sample_name,
            'sample_sex': self.sample_sex,
            'sample_group': self.sample_group,
            'sample_type': self.sample_type,
            'pileup_data': pileup_data,
            'region_counts': region_counts
        }

        pickle.dump(sample_data, open(self.extraction_file, "wb"))

    def load_from_file(self, extraction_file=None):

        if extraction_file is not None:
            self.extraction_file = extraction_file

        assert self.extraction_file is not None, 'Extraction file path is None.'
        assert os.path.exists(self.extraction_file), 'Extraction file does not exist: {}'.format(
            self.extraction_file)

        sample_data = pickle.load(open(self.extraction_file, "rb"))

        region_counts = None
        if sample_data.get('region_counts') is not None:
            region_counts = pd.DataFrame(
                sample_data['region_counts'], dtype=object)

        self.pileup = pd.DataFrame(sample_data['pileup_data'])
        self.sample_bam = sample_data['sample_bam']
        self.sample_name = sample_data['sample_name']
        self.sample_sex = sample_data['sample_sex']
        self.sample_group = sample_data['sample_group']
        self.sample_type = sample_data['sample_type']
        self.region_counts = region_counts



import logging


def get_logger(debug=False):
    FORMAT = '%(levelname)s - %(asctime)-15s: %(message)s'
    logging.basicConfig(format=FORMAT)
    logger = logging.getLogger("biometrics")

    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    return logger


def standardize_sex_nomenclature(val):

    if val is None:
        return None

    # Potential inputs
    female = ['female', 'f', 'Female', 'F']
    male = ['Male', 'M', 'male', 'm']

    if val in female:
        return 'F'
    elif val in male:
        return 'M'

    return None



#!/usr/bin/env python

"""The setup script."""

import os

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()


def req_file(filename):
    """
    We're using a requirements.txt file so that pyup.io can use this for security checks
    :param filename:
    :return str:
    """
    with open(filename) as f:
        content = f.readlines()
        content = filter(lambda x: not x.startswith("#"), content)
    return [x.strip() for x in content]

with open(os.path.join(os.path.dirname(__file__), "biometrics/VERSION"), "r") as fh:
    __version__ = fh.read().strip()


setup(
    author="Charlie Murphy",
    author_email='murphyc4@mskcc.org',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    description="Package to generate sample based biometrics.",
    entry_points={
        'console_scripts': [
            'biometrics=biometrics.cli:main',
        ],
    },
    install_requires=req_file("requirements.txt"),
    license="Apache Software License 2.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='biometrics',
    name='biometrics',
    packages=find_packages(include=['biometrics', 'biometrics.*']),
    package_data={
        "": ['requirements.txt', 'requirements_dev.txt'],
    },
    test_suite='tests',
    url='https://github.com/msk-access/biometrics',
    version=__version__,
    zip_safe=False,
)
